import { Connection } from '@salesforce/core-bundle';
import { RegistryAccess } from '../../registry';
/**
 * This is the local "spidering" logic for agents.  Given the API name for a Bot,
 * and either an org connection or local file system paths, resolve to the top
 * level agent metadata. E.g., Bot, BotVersion, GenAiPlanner, GenAiPlugin, and
 * GenAiFunction.
 *
 * If an org connection is provided, it will query the org for GenAiPlanner and
 * GenAiPlugin metadata associated with the Bot name.
 *
 * If no org connection but directory paths are provided, it will search those
 * directory paths for BotVersion and GenAiPlanner metadata associated with the
 * Bot name.
 *
 * @param agentMdInfo Data necessary to get agent related metadata.
 * @returns An array of metadata types and possibly metadata names (Metadata entries)
 */
export declare function resolveAgentMdEntries(agentMdInfo: {
    botName: string;
    directoryPaths?: string[];
    connection?: Connection;
    registry?: RegistryAccess;
}): Promise<string[]>;
